import json
import struct
import logging

from .property import Property
from .constants import MQTTCommands, MQTTv50

logger = logging.getLogger(__name__)

LAST_MID = 0


class Packet(object):
    __slots__ = ['cmd', 'data']

    def __init__(self, cmd, data):
        self.cmd = cmd
        self.data = data


class PackageFactory(object):
    @classmethod
    async def parse_package(cls, cmd, package):
        pass

    @classmethod
    def build_package(cls, *args, **kwargs) -> bytes:
        raise NotImplementedError

    @classmethod
    def _pack_remaining_length(cls, packet, remaining_length):
        remaining_bytes = []
        while True:
            byte = remaining_length % 128
            remaining_length = remaining_length // 128
            # If there are more digits to encode, set the top bit of this digit
            if remaining_length > 0:
                byte |= 0x80

            remaining_bytes.append(byte)
            packet.append(byte)
            if remaining_length == 0:
                # FIXME - this doesn't deal with incorrectly large payloads
                return packet

    @classmethod
    def _pack_str16(cls, packet, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        packet.extend(struct.pack("!H", len(data)))
        packet.extend(data)

    @classmethod
    def _mid_generate(cls):
        global LAST_MID
        LAST_MID += 1
        if LAST_MID == 65536:
            LAST_MID = 1
        return LAST_MID

    @classmethod
    def _build_properties_data(cls, properties_dict, protocol_version):
        if protocol_version < MQTTv50:
            return bytearray()
        data = bytearray()
        result = bytearray()
        for property_name, property_value in properties_dict.items():
            try:
                property = Property.factory(name=property_name)
            except KeyError:
                continue
            property_bytes = property.dumps(property_value)
            data.extend(property_bytes)
        result += struct.pack('!B', len(data))
        result.extend(data)
        return result


class LoginPackageFactor(PackageFactory):
    @classmethod
    def build_package(cls, client_id, username, password, clean_session, keepalive, protocol, **kwargs):
        remaining_length = 2 + len(protocol.proto_name) + 1 + 1 + 2 + 2 + len(client_id)

        connect_flags = 0
        if clean_session:
            connect_flags |= 0x02

        # TODO: does we need this?
        # if cls._will:
        #     remaining_length += 2 + len(cls._will_topic) + 2 + len(cls._will_payload)
        #     connect_flags |= 0x04 | ((cls._will_qos & 0x03) << 3) | ((cls._will_retain & 0x01) << 5)

        if username is not None:
            remaining_length += 2 + len(username)
            connect_flags |= 0x80
            if password is not None:
                connect_flags |= 0x40
                remaining_length += 2 + len(password)

        command = MQTTCommands.CONNECT
        packet = bytearray()
        packet.append(command)

        prop_bytes = cls._build_properties_data(kwargs, protocol.proto_ver)
        remaining_length += len(prop_bytes)

        cls._pack_remaining_length(packet, remaining_length)
        packet.extend(struct.pack("!H" + str(len(protocol.proto_name)) + "sBBH",
                                  len(protocol.proto_name),
                                  protocol.proto_name,
                                  protocol.proto_ver,
                                  connect_flags,
                                  keepalive))

        packet.extend(prop_bytes)

        cls._pack_str16(packet, client_id)

        if username is not None:
            cls._pack_str16(packet, username)

            if password is not None:
                cls._pack_str16(packet, password)

        return packet


class SubscribePacket(PackageFactory):
    @classmethod
    def build_package(cls, topic, qos, protocol, **kwargs) -> bytes:
        remaining_length = 2
        if not isinstance(topic, (list, tuple)):
            topics = [topic]
        else:
            topics = topic

        for t in topics:
            remaining_length += 2 + len(t) + 1

        properties = cls._build_properties_data(kwargs, protocol.proto_ver)
        remaining_length += len(properties)

        command = MQTTCommands.SUBSCRIBE | (False << 3) | 0x2
        packet = bytearray()
        packet.append(command)
        cls._pack_remaining_length(packet, remaining_length)
        local_mid = cls._mid_generate()
        packet.extend(struct.pack("!H", local_mid))
        packet.extend(properties)
        for t in topics:
            cls._pack_str16(packet, t)
            packet.append(qos)

        logger.info('[SEND SUB] %s', topics)

        return packet


class SimpleCommandPacket(PackageFactory):
    @classmethod
    def build_package(cls, command) -> bytes:
        return struct.pack('!BB', command, 0)


class PublishPacket(PackageFactory):
    @classmethod
    def build_package(cls, topic, payload, qos, retain, protocol, dup=False, **kwargs) -> bytes:
        command = MQTTCommands.PUBLISH | ((dup & 0x1) << 3) | (qos << 1) | (retain & 0x1)
        packet = bytearray()
        packet.append(command)

        if isinstance(payload, dict):
            payload = json.dumps(payload)

        if isinstance(payload, (int, float)):
            payload = str(payload).encode('ascii')
        elif isinstance(payload, str):
            payload = payload.encode()
        elif payload is None:
            payload = b''

        payload_size = len(payload)

        if payload_size > 268435455:
            raise ValueError('Payload too large.')

        remaining_length = 2 + len(topic) + payload_size
        prop_bytes = cls._build_properties_data(kwargs, protocol_version=protocol.proto_ver)
        remaining_length += len(prop_bytes)

        if payload_size == 0:
            logger.debug("Sending PUBLISH (q%d), '%s' (NULL payload)", qos, topic)
        else:
            logger.debug("Sending PUBLISH (q%d), '%s', ... (%d bytes)", qos, topic, payload_size)

        mid = cls._mid_generate()

        if qos > 0:
            # For message id
            remaining_length += 2

        cls._pack_remaining_length(packet, remaining_length)
        cls._pack_str16(packet, topic)

        if qos > 0:
            # For message id
            packet.extend(struct.pack("!H", mid))
        packet.extend(prop_bytes)

        packet.extend(payload)

        return packet


class CommandWithMidPacket(PackageFactory):

    @classmethod
    def build_package(cls, cmd, mid, dup) -> bytes:
        if dup:
            cmd |= 0x8

        remaining_length = 2
        packet = struct.pack('!BBH', cmd, remaining_length, mid)
        return packet


import asyncio
import logging
import struct

from . import package
from .constants import MQTTv50, MQTTCommands

logger = logging.getLogger(__name__)


class BaseMQTTProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, buffer_size=2**16, loop=None):
        if not loop:
            loop = asyncio.get_event_loop()

        self._connection = None
        self._transport = None

        self._connected = asyncio.Event(loop=loop)

        reader = asyncio.StreamReader(limit=buffer_size, loop=loop)
        super(BaseMQTTProtocol, self).__init__(reader, loop=loop)

    def set_connection(self, conn):
        self._connection = conn

    def _parse_packet(self):
        raise NotImplementedError

    def connection_made(self, transport: asyncio.Transport):
        super(BaseMQTTProtocol, self).connection_made(transport)

        logger.info('[CONNECTION MADE]')
        self._transport = transport

        self._connected.set()

    def data_received(self, data):
        super(BaseMQTTProtocol, self).data_received(data)

    def write_data(self, data: bytes):
        if not self._transport.is_closing():
            self._transport.write(data)
        else:
            logger.warning('[TRYING WRITE TO CLOSED SOCKET]')

    def connection_lost(self, exc):
        self._connected.clear()
        super(BaseMQTTProtocol, self).connection_lost(exc)
        if exc:
            logger.warning('[EXC: CONN LOST]', exc_info=exc)
        else:
            logger.info('[CONN CLOSE NORMALLY]')

    async def read(self, n=-1):
        bs = await self._stream_reader.read(n=n)
        if not bs:
            logger.warning('[READ] no data received, breaking connection')
            self.connection_lost(None)
        return bs



class MQTTProtocol(BaseMQTTProtocol):
    proto_name = b'MQTT'
    proto_ver = MQTTv50

    def __init__(self, *args, **kwargs):
        super(MQTTProtocol, self).__init__(*args, **kwargs)
        self._queue = asyncio.Queue()

        self._disconnect = asyncio.Event()

        self._read_loop_future = None

    def connection_made(self, transport: asyncio.Transport):
        super().connection_made(transport)
        self._read_loop_future = asyncio.ensure_future(self._read_loop())

    async def send_auth_package(self, client_id, username, password, clean_session, keepalive, **kwargs):
        pkg = package.LoginPackageFactor.build_package(client_id, username, password, clean_session,
                                                       keepalive, self, **kwargs)
        self.write_data(pkg)

    def send_subscribe_packet(self, topic, qos, **kwargs):
        pkg = package.SubscribePacket.build_package(topic, qos, self, **kwargs)
        self.write_data(pkg)

    def send_simple_command_packet(self, cmd):
        pkg = package.SimpleCommandPacket.build_package(cmd)
        self.write_data(pkg)

    def send_ping_request(self):
        self.send_simple_command_packet(MQTTCommands.PINGREQ)

    def send_publish(self, topic, payload, qos, retain, **kwargs):
        pkg = package.PublishPacket.build_package(topic, payload, qos, retain, self, **kwargs)
        self.write_data(pkg)

    def send_command_with_mid(self, cmd, mid, dup):
        pkg = package.CommandWithMidPacket.build_package(cmd, mid, dup)
        self.write_data(pkg)

    async def _read_packet(self):
        remaining_count = []
        remaining_length = 0
        remaining_mult = 1

        while True:
            byte, = struct.unpack("!B", await self.read(1))
            remaining_count.append(byte)

            if len(remaining_count) > 4:
                logger.warning('[MQTT ERR PROTO] RECV MORE THAN 4 bytes for remaining length.')
                return None

            remaining_length += (byte & 127) * remaining_mult
            remaining_mult *= 128

            if byte & 128 == 0:
                break

        packet = b''
        while remaining_length > 0:
            chunk = await self.read(remaining_length)
            remaining_length -= len(chunk)
            packet += chunk

        return packet

    async def _read_loop(self):
        await self._connected.wait()

        while self._connected.is_set():
            byte = await self.read(1)
            if not byte:
                continue
            command, = struct.unpack("!B", byte)
            packet = await self._read_packet()
            self._connection.put_package((command, packet))

    def connection_lost(self, exc):
        super(MQTTProtocol, self).connection_lost(exc)
        self._connection.put_package((MQTTCommands.DISCONNECT, b''))

        self._read_loop_future.cancel()
        self._read_loop_future = None

        self._queue = asyncio.Queue()



import enum

import logging

# Message types

class MQTTCommands(enum.IntEnum):
    CONNECT = 0x10
    CONNACK = 0x20
    PUBLISH = 0x30
    PUBACK = 0x40
    PUBREC = 0x50
    PUBREL = 0x60
    PUBCOMP = 0x70
    SUBSCRIBE = 0x80
    SUBACK = 0x90
    UNSUBSCRIBE = 0xA0
    UNSUBACK = 0xB0
    PINGREQ = 0xC0
    PINGRESP = 0xD0
    DISCONNECT = 0xE0

# CONNACK codes
CONNACK_ACCEPTED = 0
CONNACK_REFUSED_PROTOCOL_VERSION = 1
CONNACK_REFUSED_IDENTIFIER_REJECTED = 2
CONNACK_REFUSED_SERVER_UNAVAILABLE = 3
CONNACK_REFUSED_BAD_USERNAME_PASSWORD = 4
CONNACK_REFUSED_NOT_AUTHORIZED = 5


# Error values
MQTT_ERR_AGAIN = -1
MQTT_ERR_SUCCESS = 0
MQTT_ERR_NOMEM = 1
MQTT_ERR_PROTOCOL = 2
MQTT_ERR_INVAL = 3
MQTT_ERR_NO_CONN = 4
MQTT_ERR_CONN_REFUSED = 5
MQTT_ERR_NOT_FOUND = 6
MQTT_ERR_CONN_LOST = 7
MQTT_ERR_TLS = 8
MQTT_ERR_PAYLOAD_SIZE = 9
MQTT_ERR_NOT_SUPPORTED = 10
MQTT_ERR_AUTH = 11
MQTT_ERR_ACL_DENIED = 12
MQTT_ERR_UNKNOWN = 13
MQTT_ERR_ERRNO = 14
MQTT_ERR_QUEUE_SIZE = 15


# MQTT protocol versions
MQTTv311 = 4
MQTTv50 = 5



import datetime

from .client import Client
from gmqtt.mqtt.protocol import BaseMQTTProtocol

__author__ = "Mikhail Turchunovich"
__email__ = 'mitu@gurtam.com'
__copyright__ = ("Copyright 2013-%d, Gurtam; " % datetime.datetime.now().year,)

__credits__ = [
    "Mikhail Turchunovich",
    "Elena Nikolaichik"
]
__version__ = "0.0.14"


__all__ = [
    'Client', 'BaseMQTTProtocol'
]



import struct


class Property:
    def __init__(self, id_, bytes_struct, name, allowed_packages):
        self.id = id_
        self.bytes_struct = bytes_struct
        self.name = name
        self.allowed_packages = allowed_packages

    def loads(self, bytes_array):
        # returns dict with property name-value and remaining bytes which do not belong to this property
        if self.bytes_struct == 'u8':
            # First two bytes in UTF-8 encoded properties correspond to unicode string length
            str_len, = struct.unpack('!H', bytes_array[:2])
            value = bytes_array[2:2+str_len].decode('utf-8')
            left_str = bytes_array[2+str_len:]

        else:
            value, left_str = self.unpack_helper(self.bytes_struct, bytes_array)
        return {self.name: value}, left_str

    def unpack_helper(self, fmt, data):
        # unpacks property value according to format, returns value and remaining bytes
        size = struct.calcsize(fmt)
        value = struct.unpack(fmt, data[:size])
        left_str = data[size:]
        if len(value) == 1:
            value = value[0]
        return value, left_str

    def dumps(self, data):
        # packs property value into byte array
        packet = bytearray()
        packet.extend(struct.pack('!B', self.id))
        if self.bytes_struct == 'u8':
            if isinstance(data, str):
                data = data.encode('utf-8')
            packet.extend(struct.pack("!H", len(data)))
            packet.extend(data)
            return packet
        packet.extend(struct.pack(self.bytes_struct, data))
        return packet


    @classmethod
    def factory(cls, id_=None, name=None):
        if (name is None and id_ is None) or (name is not None and id_ is not None):
            raise ValueError('Either id or name should be not None')
        if name is not None:
            return PROPERTIES_BY_NAME[name]
        else:
            return PROPERTIES_BY_ID[id_]


PROPERTIES = [
    Property(1, '!B', 'payload_format_id', ['PUBLISH', ]),
    Property(2, '!L', 'message_expiry_interval', ['PUBLISH', ]),
    Property(8, 'u8', 'response_topic', ['PUBLISH', ]),
    Property(17, '!L', 'session_expiry_interval', ['CONNECT', ]),
    Property(24, '!L', 'will_delay_interval', ['CONNECT', ]),
    Property(18, 'u8', 'client_id', ['CONNACK', ]),
    Property(33, '!H', 'receive_maximum', ['CONNECT', ]),
    Property(34, '!H', 'topic_alias_maximum', ['CONNECT', 'CONNACK']),
    Property(35, '!H', 'topic_alias', ['CONNECT', 'CONNACK']),
    Property(36, '!B', 'max_qos', ['CONNACK', ]),
    Property(39, '!L', 'maximum_packet_size', ['CONNECT', 'CONNACK']),
    Property(41, '!B', 'sub_id_available', ['CONNACK', ]),
]

PROPERTIES_BY_ID = {pr.id: pr for pr in PROPERTIES}
PROPERTIES_BY_NAME = {pr.name: pr for pr in PROPERTIES}





import asyncio
import uuid

from .mqtt.protocol import MQTTProtocol
from .mqtt.connection import MQTTConnection
from .mqtt.handler import MqttPackageHandler
from .mqtt.constants import MQTTv311, MQTTv50


class Client(MqttPackageHandler):
    def __init__(self, client_id, clean_session=True, transport='tcp'):
        super(Client, self).__init__()
        self._client_id = client_id or uuid.uuid4().hex

        self._clean_session = clean_session
        self._transport = transport

        self._connection = None

        self._username = None
        self._password = None

        self._host = None
        self._port = None

    def set_auth_credentials(self, username, password=None):
        self._username = username.encode()
        self._password = password
        if isinstance(self._password, str):
            self._password = password.encode()

    async def connect(self, host, port=1883, clean_session=True, keepalive=60, version=MQTTv50):
        # Init connection
        self._host = host
        self._port = port

        MQTTProtocol.proto_ver = version

        self._connection = await self._create_connection(host, port=self._port, clean_session=clean_session, keepalive=keepalive)

        await self._connection.auth(self._client_id, self._username, self._password)
        await self._connected.wait()

        if self._error:
            raise self._error

    async def _create_connection(self, host, port, clean_session, keepalive):
        self._reconnect = True
        connection = await MQTTConnection.create_connection(host, port, clean_session, keepalive)
        connection.set_handler(self)
        return connection

    async def reconnect(self):
        await self.disconnect()
        await asyncio.sleep(1)
        self._connection = await self._create_connection(self._host, self._port, clean_session=True, keepalive=60)
        await self._connection.auth(self._client_id, self._username, self._password)

    async def disconnect(self):
        self._reconnect = False
        if self._connection:
            await self._connection.close()

    def subscribe(self, topic, qos=0, **kwargs):
        self._connection.subsribe(topic, qos, **kwargs)

    def publish(self, topic, payload, qos=0, retain=False, **kwargs):
        self._connection.publish(topic, payload, qos=qos, retain=retain, **kwargs)

    def _send_simple_command(self, cmd):
        self._connection.send_simple_command(cmd)

    def _send_command_with_mid(self, cmd, mid, dup):
        self._connection.send_command_with_mid(cmd, mid, dup)

    @property
    def protocol_version(self):
        return self._connection._protocol.proto_ver \
            if self._connection is not None else MQTTv50
